/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "fggl/phys/bullet/types.hpp"

namespace fggl::phys::bullet {

	BulletPhysicsEngine::BulletPhysicsEngine(ecs3::World* world) : m_ecs(world), m_config(), m_world(nullptr) {
		m_config.broadphase = new btDbvtBroadphase();
		m_config.collisionConfiguration = new btDefaultCollisionConfiguration();
		m_config.dispatcher = new btCollisionDispatcher(m_config.collisionConfiguration);
		m_config.solver = new btSequentialImpulseConstraintSolver();

		m_world = new btDiscreteDynamicsWorld(
			m_config.dispatcher,
			m_config.broadphase,
			m_config.solver,
			m_config.collisionConfiguration);

		m_debug = std::make_unique<debug::BulletDebugDrawList>();
		m_world->setDebugDrawer( m_debug.get() );
		m_debug->setDebugMode(1);

		m_ecs->addDeathListener( [this](auto entity) { this->onEntityDeath(entity);} );
	}

	void BulletPhysicsEngine::step() {
		checkForPhys();

		auto dynamicEnts = m_ecs->findMatching<phys::Dynamics>();
		for (auto& ent : dynamicEnts) {
			auto* dynamicComp = m_ecs->get<phys::Dynamics>(ent);
			auto* bulletProxy = m_ecs->get<phys::bullet::BulletBody>(ent);
			if ( bulletProxy == nullptr ) {
				std::cerr << "entity has dynamics but isn't a physics object!" << std::endl;
				continue;
			}

			const auto forceVec = dynamicComp->force;
			if ( glm::length(forceVec) != 0.F ) {
				bulletProxy->body->applyCentralForce({forceVec.x, forceVec.y, forceVec.z});
				bulletProxy->body->activate();
				dynamicComp->force = math::VEC3_ZERO;
			}
		}

		m_world->stepSimulation(60.0f);
		syncToECS();
		dealWithCollisions();

		m_world->debugDrawWorld();
	}


	static void build_motation_state(math::Transform* myState, BulletBody* btState) {
		auto myPos = myState->origin();
		btVector3 position{myPos.x, myPos.y, myPos.z};

		auto myRot = myState->euler();
		btQuaternion rotation;
		rotation.setEulerZYX(myRot.x, myRot.y, myRot.z);

		btTransform transform;
		transform.setIdentity();
		transform.setRotation(rotation);
		transform.setOrigin(position);

		btState->motion = new btDefaultMotionState(transform);
	}

	inline btCollisionShape* shapeToBullet(const phys::RigidBody* fgglBody) {
		if ( fgglBody->shape == nullptr ) {
			// they forgot to put a shape, we'll assume a unit sphere to avoid crashes...
			return new btSphereShape(0.5f);
		}

		switch (fgglBody->shape->type) {
			case phys::ShapeType::BOX: {
				auto extents = ((phys::Box*)fgglBody->shape)->extents;
				return new btBoxShape({extents.x, extents.y, extents.z});
			}
			case phys::ShapeType::SPHERE: {
				float radius = ((phys::Sphere*)fgglBody->shape)->radius;
				return new btSphereShape(radius);
			}
			default:
				return new btSphereShape(0.5f);
		}
	}

	void BulletPhysicsEngine::checkForPhys() {
		// FIXME without reactive-based approaches this is very slow
		auto entsWantPhys = m_ecs->findMatching<phys::RigidBody>();
		for (auto ent : entsWantPhys) {
			auto* btComp = m_ecs->get<BulletBody>(ent);
			if ( btComp != nullptr ) {
				// they are already in the simluation
				continue;
			}

			btComp = m_ecs->add<BulletBody>(ent);
			const auto* fgBody = m_ecs->get<phys::RigidBody>(ent);

			// setup the starting motion state
			auto* transform = m_ecs->get<math::Transform>(ent);
			build_motation_state(transform, btComp);

			btCollisionShape* colShape = shapeToBullet(fgBody);
			btVector3 localInt(0, 0, 0);
			if ( !fgBody->isStatic() ) {
				colShape->calculateLocalInertia(fgBody->mass, localInt);
			}

			// setup the construction information
			btRigidBody::btRigidBodyConstructionInfo bodyCI(
				fgBody->mass,
				btComp->motion,
				colShape,
				localInt
			);
			auto* body = new btRigidBody(bodyCI);
			body->setUserIndex(static_cast<int>(ent));

			btComp->body = body;

			if (!fgBody->isStatic()) {
				body->setRestitution(0.5f);
				body->setRollingFriction(0.0f);
			}

			if ( fgBody->type == BodyType::KINEMATIC ) {
				body->setCollisionFlags( body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT );
			}

			m_world->addRigidBody(btComp->body );
		}
	}

	void BulletPhysicsEngine::syncToECS() {
		auto physEnts = m_ecs->findMatching<BulletBody>();
		for (auto& ent : physEnts) {
			auto* transform = m_ecs->get<math::Transform>(ent);
			auto* physBody = m_ecs->get<BulletBody>(ent);

			btTransform bTransform;
			physBody->motion->getWorldTransform(bTransform);

			// set calculate and set ecs position
			auto& btPos = bTransform.getOrigin();
			math::vec3 pos{btPos.x(), btPos.y(), btPos.z()};
			transform->origin( pos );

			// set the rotation
			const auto& btRot = bTransform.getRotation();
			math::quat rot{btRot.x(), btRot.y(), btRot.z(), btRot.w()};
			transform->setRotation(rot);
		}
	}

	BulletPhysicsEngine::~BulletPhysicsEngine() {
		// clean up the rigid bodies
		auto entRB = m_ecs->findMatching<BulletBody>();
		for (auto& ent : entRB) {
			auto* bulletBody = m_ecs->get<BulletBody>(ent);
			delete bulletBody->motion;
			m_world->removeCollisionObject(bulletBody->body);
			delete bulletBody->body;
			m_ecs->remove<BulletBody>(ent);
		}

		// delete the world
		delete m_world;

		// cleanup the configuration object
		delete m_config.solver;
		delete m_config.broadphase;
		delete m_config.dispatcher;
		delete m_config.collisionConfiguration;
	}

	static void handleCollisionCallbacks(ecs3::World* world, ecs3::entity_t owner, ecs3::entity_t other) {
		auto* callbacks = world->tryGet<CollisionCallbacks>(owner);
		if ( callbacks == nullptr ) {
			return;
		}

		auto* cache = world->tryGet<CollisionCache>(owner);
		if ( cache != nullptr ) {
			auto itr = cache->collisions.find(other);
			if ( itr == cache->collisions.end() ) {
				if ( callbacks->onEnter != nullptr ) {
					callbacks->onEnter(owner, other);
				}
				cache->collisions.insert(  other );
			} else {
				if ( callbacks->onStay != nullptr ) {
					callbacks->onStay(owner, other);
				}
			}
		}
	}

	void BulletPhysicsEngine::dealWithCollisions() {
		// flush collision caches
		auto caches = m_ecs->findMatching<CollisionCache>();
		for( auto& ent : caches) {
			auto* cache = m_ecs->get<CollisionCache>(ent);
			std::swap(cache->collisions, cache->lastFrame);
			cache->collisions.clear();
		}

		// deal with the number of manifolds
		const auto numManifolds = m_config.dispatcher->getNumManifolds();
		for ( int i=0; i< numManifolds; ++i) {
			auto* contactManifold = m_config.dispatcher->getManifoldByIndexInternal(i);

			const auto* body0 = contactManifold->getBody0();
			const auto* body1 = contactManifold->getBody1();

			// FIXME rigid body didn't die according to plan!
			if ( body0->getUserIndex() == ecs::NULL_ENTITY || body1->getUserIndex() == ecs::NULL_ENTITY) {
				continue;
			}

			if ( !m_ecs->alive( body0->getUserIndex()) || !m_ecs->alive(body1->getUserIndex()) ) {
				continue;
			}

			std::cerr << "contact: " << body0->getUserIndex() << " on " << body1->getUserIndex() << std::endl;

			handleCollisionCallbacks(m_ecs, body0->getUserIndex(), body1->getUserIndex());
			handleCollisionCallbacks(m_ecs, body1->getUserIndex(), body0->getUserIndex());
		}

		// note conacts that have ended
		caches = m_ecs->findMatching<CollisionCache>(); // re-fetch, entities can (and probably do) die in handlers.
		for( auto& ent : caches) {
			auto* cache = m_ecs->get<CollisionCache>(ent);

			// if we don't have an exit callback, it's pointless checking the cache
			auto* callbacks = m_ecs->tryGet<CollisionCallbacks>(ent);
			if ( callbacks == nullptr || callbacks->onExit == nullptr){
				continue;
			}

			// check if a collision has ended
			for (const auto& other : cache->lastFrame) {
				auto itr = cache->collisions.find(other);
				if (itr == cache->collisions.end()) {
					callbacks->onExit( ent,other);
				}
			}
		}
	}

	void BulletPhysicsEngine::onEntityDeath(ecs::entity_t entity)  {
			auto* btPhysics = m_ecs->tryGet<BulletBody>(entity);
			if ( btPhysics != nullptr) {
				btPhysics->body->setUserIndex( ecs::NULL_ENTITY );
				m_world->removeRigidBody( btPhysics->body );
			}
	}

} // namespace fggl::phys::bullet